#include <Python.h>
#include <numpy/arrayobject.h>
#include <dlfcn.h>
#include <omp.h>

void init_py_delen(int rank);

PyObject* fill_py_list(const int lsize, const double *arr);

void  get_delensed_theory(const int rank, const int cls_size, const int filt_size, const double *cltt_unlen, const double *clte_unlen, const double *clee_unlen, const double *lpp_th, const double *clpp, const double *L, const double *NL, const double kappa_lmin, const double kappa_lmax, double *cltt_delen, double *clte_delen, double *clee_delen);


void init_py_delen(int rank)
{
	if (!Py_IsInitialized())
    {
    	Py_Initialize();   /* initialize python interpreter */
	
		/* import camb here first while initializing */
    	PyObject *camb; 
		const char camb_name[] = "camb";
		camb = PyImport_ImportModule(camb_name);
		/* check for errors */
    	if(camb==NULL) 
		{
			if (PyErr_Occurred()) PyErr_Print(); 
			printf("error importing camb\n");
		}
		Py_XDECREF(camb);
	}

	return;
}


PyObject* fill_py_list(const int lsize, const double *arr)
{
	int i;
	PyObject *pList; /* new python list object */

	if(!Py_IsInitialized()) Py_Initialize();   /* initialize python interpreter */

	/* create a new list and check for errors*/
	pList = PyList_New(lsize);
    if (pList == NULL)
    {
        printf("Error creating a new python list.\n");
        if (PyErr_Occurred()) PyErr_Print();
        exit(1);
    }

	/* fill the list with the contents of the input array */
    for (i=0; i<lsize; i++)
    {
        PyList_SetItem(pList, i, PyFloat_FromDouble(arr[i]));
    }

	return pList;
}


void  get_delensed_theory(const int rank, const int cls_size, const int filt_size, const double *cltt_unlen, const double *clte_unlen, const double *clee_unlen, const double *lpp_th, const double *clpp, const double *L, const double *NL, const double kappa_lmin, const double kappa_lmax, double *cltt_delen, double *clte_delen, double *clee_delen)
{
    /* use the PyObject type to declare variables (which are pointers?) */
    PyObject *pName;   /* name of the .py file? */
    PyObject *pModule; /* holds the imported module? */
    PyObject *pFunc;   /* holds the function */
    PyObject *pArgs;   /* python tuple of arguments to pass to function */
    PyObject *pValue;  /* result of the function */
    char pname_fname[] = "actpol_delensing_py";   /* file name of .py file */
    char pfunc_name[]  = "get_delensed";          /* name of function in .py file */
	double  y_val;     /* tmp to hold value of array elements returned by python */
	void* np_ptr;      /* needed to get value from numpy array? */
	PyObject* x_val;   /* tmp to hold individual array element that is passed to python */
	PyObject *ret;     /* hold item of returned tuple from python */
	PyObject *cltt_list, *clte_list, *clee_list, *lpp_list, *clpp_list, *L_list, *NL_list; /* need to pass python lists to python func */
	int i;	 


	/* make sure python is initialized */
    if(!Py_IsInitialized())
	{
		printf("rank %d: initializing python\n", rank);
		init_py_delen(rank);
		if (PyErr_Occurred()) PyErr_Print();
	}

	/* import delensing module */
    pName = PyUnicode_FromString(pname_fname);  /* get python string for name */
    if (pName == NULL)  /* check for errors */
    {
    	printf("Problem getting pName");
    	if (PyErr_Occurred()) PyErr_Print();
    }
	pModule = PyImport_Import(pName);  /* import python module */
    Py_DECREF(pName); 
	

    /* want to raise an error if something went wrong, otherwise use the function */
    if (pModule != NULL)   /* everything is ok so far */
    {
		/* get the function from the module */	
		if (!PyObject_HasAttrString(pModule, pfunc_name))
		{
			printf("module doesn't have the function\n");
			if (PyErr_Occurred()) PyErr_Print();
			exit(1);
		}
        pFunc = PyObject_GetAttrString(pModule, pfunc_name);

        if (pFunc && PyCallable_Check(pFunc))  /* check if function is callable */
        {
            pArgs = PyTuple_New(9);  /* create tuple to hold arguments */
			if (pArgs==NULL)         /* check for errors */
			{
				printf("Error creating python tuple to hold arguments\n");
				if (PyErr_Occurred()) PyErr_Print();
				exit(1);
			}

            /* set the arguments */
			/* get python lists for unlensed spectra, ells, lensing potential, etc. */
			cltt_list = fill_py_list(cls_size, cltt_unlen);
            PyTuple_SetItem(pArgs, 0, cltt_list);

			clte_list = fill_py_list(cls_size, clte_unlen);
			PyTuple_SetItem(pArgs, 1, clte_list);

			clee_list = fill_py_list(cls_size, clee_unlen);
			PyTuple_SetItem(pArgs, 2, clee_list);

			lpp_list = fill_py_list(cls_size, lpp_th);
			PyTuple_SetItem(pArgs, 3, lpp_list);

			clpp_list = fill_py_list(cls_size, clpp);
			PyTuple_SetItem(pArgs, 4, clpp_list);

			L_list = fill_py_list(filt_size, L);
			PyTuple_SetItem(pArgs, 5, L_list);

			NL_list = fill_py_list(filt_size, NL);
			PyTuple_SetItem(pArgs, 6, NL_list);

			PyTuple_SetItem(pArgs, 7, PyFloat_FromDouble(kappa_lmin));
			PyTuple_SetItem(pArgs, 8, PyFloat_FromDouble(kappa_lmax));

            /* call the function and store the result in pValue */
            pValue = PyObject_CallObject(pFunc, pArgs);
			if (PyErr_Occurred()) PyErr_Print();

            Py_DECREF(pArgs);  /* done with pArgs */

            /* if something went wrong, raise an error */
            if (pValue != NULL)
            {
				/* ok so far; get a C array from the returned python numpy array */

				/* delensed tt */
				ret = PyTuple_GetItem(pValue, 0);
				if(ret==NULL)
				{
					printf("Error getting result from python\n");
					if (PyErr_Occurred()) PyErr_Print();
					exit(1);
				}
				for (i=0; i<cls_size; i++)
				{
					/* get a pointer to np array element i; then get the python float from the np array, 
					   and convert it to a C double, and add the value to the C array. */
					np_ptr = PyArray_GETPTR1(ret, i);  
					x_val = PyArray_GETITEM(ret, np_ptr);
					if(x_val==NULL) 
					{
						printf("error getting array element from array\n");
						if (PyErr_Occurred()) PyErr_Print();
						exit(1);
					}
					y_val = PyFloat_AsDouble(x_val);
					if(y_val==-1.0)
					{
						printf("checking for error converting python float to C double\n");
						if (PyErr_Occurred()) {PyErr_Print(); exit(1);}
					}
					cltt_delen[i] = y_val;
				}

				/* delensed te */
				ret = PyTuple_GetItem(pValue, 1);
				if(ret==NULL)
                {
                    printf("Error getting result from python\n");
                    if (PyErr_Occurred()) PyErr_Print();
                    exit(1);
                }
				for (i=0; i<cls_size; i++)
				{
					np_ptr = PyArray_GETPTR1(ret, i);
					y_val = PyFloat_AsDouble(PyArray_GETITEM(ret, np_ptr));
					clte_delen[i] = y_val;
				}

				/* delensed ee */
				ret = PyTuple_GetItem(pValue, 2);
				if(ret==NULL)
                {
                    printf("Error getting result from python\n");
                    if (PyErr_Occurred()) PyErr_Print();
                    exit(1);
                }
				for (i=0; i<cls_size; i++)
				{
					np_ptr = PyArray_GETPTR1(ret, i);
					y_val = PyFloat_AsDouble(PyArray_GETITEM(ret, np_ptr));
					clee_delen[i] = y_val;
				}

                Py_DECREF(pValue);
				Py_XDECREF(x_val);
            }

            else  /* something went wrong */
            {
				printf("result of python function was NULL \n");
                Py_DECREF(pFunc);
                Py_DECREF(pModule);
                PyErr_Print();  /* print the traceback of the error */
                exit(1);
            }
        }

        else   /* something went wrong when getting the function */
        {
            if (PyErr_Occurred()) PyErr_Print();
            printf("Cannot find the function \'%s\'\n", pfunc_name);
        }

        /* done using pFunc and pModule */
        Py_XDECREF(pFunc);
        Py_DECREF(pModule);
    }
    else  /* error while loading module */
    {
        PyErr_Print();
        printf("Failed to load \'%s\'\n", pname_fname);
        exit(1);
    }

    return;
}

