module actpol_likelihood

  use iso_c_binding
  use omp_lib

  ! cosmoMC modules
  use MatrixUtils
  use settings
  use CosmologyTypes
  use CosmoTheory
  use Calculator_Cosmology
  use Likelihood_Cosmology
  use MpiUtils
  ! ACTPol modules
  use Options
  use ACTPol_like
  use ACTPol_delens


  implicit none

  real(kind=c_double), dimension(:), allocatable, target :: L_vals, NL  ! noise filter

  ! interface with cosmoMC: 
  type, extends(TCosmoCalcLikelihood) :: ACTPolLikelihood

    ! directories holding the data
    ! data_dir is main directory, act_data_dir is for actpol data
    ! NOTE: assumed data_dir contains act_data_dir and directory named 'Fg'
    ! files with spec_fname, etc. are inside act_data_dir
    character(len=240) :: data_dir, act_data_dir
    ! filenames:
    character(len=240) :: spec_fname, cov_fname    ! spectra (as a single data vector) and covmat filenames
    character(len=240) :: bbltt_fname, bblte_fname, bblee_fname ! binning matrices for TT, TE, EE spectra 
    ! location and name of .paramnames file:
    character(len=240) :: paramnames_path

    ! other settings
    integer :: tt_lmax                    ! lmax for CMB theory (above this is only FG)
    integer :: nbintt, nbinte, nbinee     ! number of bins for TT, TE, and EE spectra
    integer :: nbint                      ! total numbers of bins in data vector
    integer :: lmax_win                   ! lmax of full window function
    integer :: bmax0                      ! total # bins in full window func
    integer :: b1tt, b1te, b1ee           ! first bin used in TT/TE/EE window func
    logical :: use_cls                    ! if true, data is in cls; will convert theory from dls to cls
    logical :: use_tt, use_te, use_ee     ! do we want to use temp+polarization?
    logical :: invcov                     ! if True, use inverse covmat as input

    ! settings for data only (not sims)
    logical :: data_run                   ! True if running on data, otherwise false
    character(len=240) :: leakage_correction_fname ! filename for leaking correction; applied to the final binned theory

    ! settings for testing purposes
    logical :: save_test_output           ! for testing; if true, save various intermediate steps
    real(mcp) :: expected_like            ! for testing; expected likelihood value - needed if save_test_output=T
    character(len=240) :: test_save_dir, test_save_root  ! only if save_test_output=T

    ! delensing
    logical :: do_delensing               ! do we want to get the delensed theory?
    character(len=240) :: filt_fname      ! filename for noise filter
    integer :: filt_lmax, flmax           ! Lmax for noise filter
    integer :: theo_lmax                  ! lmax for CMB theory from CAMB/cosmoMC
    real(kind=c_double) :: kappa_lmin, kappa_lmax ! lmin, lmax used for filter
    integer(kind=c_int) :: rank           ! mpi rank 

    contains
      procedure :: LogLike => ACTPol_LnLike
      procedure :: actpol_like_init  

  end type ACTPolLikelihood


  contains

    ! subroutine to interface with cosmoMC: called once from DataLikelihoods.f90 at beginning
    subroutine ACTPolLikelihood_Add(LikeList, Ini)
      class(TLikelihoodList) :: Likelist
      class(TSettingIni) :: ini
      type(ACTPolLikelihood), pointer :: this

      integer :: lun
      logical :: ok

      ! read from ini file and assign values to variables defined in ACTPolLikelihood
      ! the second argument in the Ini%Read functions gives the default value
      if (Ini%Read_Logical('use_actpol',.false.)) then
        allocate(this) 
        this%data_dir = Ini%Read_String('data_dir_path', NotFoundFail=.true.) 
        this%act_data_dir = Ini%Read_String('act_data_dir', NotFoundFail=.true.)
        this%spec_fname = Ini%Read_String('spec_fname', NotFoundFail=.true.)
        this%cov_fname = Ini%Read_String('cov_fname', NotFoundFail=.true.) 
        this%paramnames_path = Ini%Read_String('paramnames_path', NotFoundFail=.true.)

        this%bbltt_fname = Ini%Read_String('bbltt_fname', NotFoundFail=.true.)
        this%bblte_fname = Ini%Read_String('bblte_fname', NotFoundFail=.true.)
        this%bblee_fname = Ini%Read_String('bblee_fname', NotFoundFail=.true.)

        this%tt_lmax = Ini%Read_Int('tt_ell_max', 5300)
        this%lmax_win = Ini%Read_Int('lmax_win', 7925)

        this%nbintt = Ini%Read_Int('nbintt', 47)
        this%nbinte = Ini%Read_Int('nbinte', 49)
        this%nbinee = Ini%Read_Int('nbinee', 49)
        this%nbint = nspectt*this%nbintt + nspecte*this%nbinte + nspecee*this%nbinee!Ini%Read_Int('nbint', 484)
        this%bmax0 = Ini%Read_Int('bmax0', 59)
        this%b1tt = Ini%Read_Int('b1tt', 13)
        this%b1te = Ini%Read_Int('b1te', 11)
        this%b1ee = Ini%Read_Int('b1ee', 11)

        this%use_cls = Ini%Read_Logical('use_cls', .false.)
        this%use_tt = Ini%Read_Logical('use_tt', .true.)
        this%use_te = Ini%Read_Logical('use_te', .true.)
        this%use_ee = Ini%Read_Logical('use_ee', .true.)
        this%invcov = Ini%Read_Logical('invcov', .false.)

        ! for data runs
        this%data_run = Ini%Read_Logical('data_run', .true.)
        this%leakage_correction_fname = " " ! if data_run = F, just pass an empty string
        if (this%data_run) then
          this%leakage_correction_fname = Ini%Read_String('leakage_correction_fname', NotFoundFail=.true.)
        end if

        ! for delensing
        this%theo_lmax = Ini%Read_Int('lmin_computed_cl', 10000)
        this%do_delensing = Ini%Read_Logical('delensing', .false.)
        if (this%do_delensing) then
          this%filt_fname = Ini%Read_String('filter_file', NotFoundFail=.true.)
          this%filt_lmax = Ini%Read_Int('filt_lmax', 9998)
          this%kappa_lmin = Ini%Read_Double('kappa_lmin', 80.d0)
          this%kappa_lmax = Ini%Read_Double('kappa_lmax', 3000.d0)
          ! determine lmax for noise filter array
          if (this%theo_lmax > this%filt_lmax) then
           this%flmax = this%filt_lmax
          else
           this%flmax = this%theo_lmax
          end if
        end if

        this%rank = GetMpiRank()

        ! save or print some intermediate steps
        this%save_test_output = Ini%Read_Logical('save_test_output', .false.)
        this%test_save_dir = " " ! if save_test_output = F, don't need save_dir or save_root
        this%test_save_root = " "
        if (this%save_test_output) then 
          this%expected_like = Ini%Read_Double('expected_like')
          this%test_save_dir = Ini%Read_String('test_save_dir', NotFoundFail=.true.)
          this%test_save_root = Ini%Read_String('test_save_root', NotFoundFail=.true.)
        end if

        ! if delensing, initialize python
        if (this%do_delensing) then
!$omp barrier
!$omp master          
          write(*,*) "rank ", this%rank, ": actpol_likelihood.f90: initializing python"
          call init_py(this%rank)
          write(*,*) "rank ", this%rank, ": actpol_likelihood.f90: python initialized"
!$omp end master
!$omp barrier
        end if
      
        call this%actpol_like_init  ! initialize likelihood: load bandpowers, binning, covmat
        call this%loadParamNames(trim(this%paramnames_path)) ! load paramnames file
        call LikeList%Add(this)   ! add to list of likelihoods in cosmoMC
      end if 

    end subroutine ACTPolLikelihood_Add


    subroutine actpol_like_init(this)
      ! initialize likelihood with info from ini file
      class(ACTPolLikelihood) :: this
      integer :: lun=72, i

      ! call initialization subroutine
      call ACTPol_likelihood_init(this%spec_fname, this%cov_fname, this%bbltt_fname, this%bblte_fname, this%bblee_fname, this%data_dir, this%act_data_dir, this%nbint, this%lmax_win, this%bmax0, this%b1tt, this%b1te, this%b1ee, this%data_run, this%leakage_correction_fname)

      if (this%do_delensing) then  ! read in filter for delensing
        allocate(L_vals(1:this%flmax+1))
        allocate(NL(1:this%flmax+1))
        open(unit=lun, file=this%filt_fname, form='formatted', status='unknown', action='read')
         do i = 1, this%flmax + 1
          read(lun,*) L_vals(i), NL(i)
         end do
        close(lun)
      end if

    end subroutine actpol_like_init


    ! get ln like function
    function ACTPol_LnLike(this, CMB, Theory, DataParams)
      class(ACTPolLikelihood) :: this
      Class(CMBParams) CMB
      Class(TCosmoTheoryPredictions), target :: Theory
      real(mcp) :: DataParams(:)
      real(mcp) :: ACTPol_LnLike

      ! foreground and nuisance params
      real(8) :: a_tsz, a_ksz, xi, a_d, a_s, a_c, a_g, a_tps, a_ps, a_gte, a_gee, a_ste, a_see, beta_c
      real(8) :: ct1, ct2, yp1, yp2

      ! variables to hold the theory Cl's
      real(kind=c_double), dimension(:), allocatable, target :: cltt_theo, clte_theo, clee_theo, clpp_theo, ells ! theory Cls
      real(kind=c_double), dimension(:), allocatable, target :: tt_delen, te_delen, ee_delen ! delensed theory 
      real(8), dimension(:,:), allocatable :: cls_vector ! vector holding theory cls that is passed to likelihood 

      ! other helpful variables
      integer :: l, lun
      real(8) :: lfact, multfact, actpol_like
      integer(kind=c_int) :: cls_size, filt_size
      character(len=240) :: cmb_theo_fname, clpp_theo_fname

      write(*,*) "rank ", this%rank, ": actpol_likelihood.f90: calculating likelihood..."

      ! read in theory Cls, and set them to zero for l = 1, 2
      allocate(cltt_theo(1:this%theo_lmax))
      allocate(clte_theo(1:this%theo_lmax))
      allocate(clee_theo(1:this%theo_lmax))
      if (this%do_delensing) then
        allocate(ells(1:this%theo_lmax))
        allocate(clpp_theo(1:this%theo_lmax))
        allocate(tt_delen(1:this%theo_lmax))
        allocate(te_delen(1:this%theo_lmax))
        allocate(ee_delen(1:this%theo_lmax))
      end if

      do l = 1, this%theo_lmax
        ! if l = 0, 1, then set cl=0
        if(l<2) then
          multfact = 0.
        else
          multfact = 1.
        end if
        lfact = multfact !* 2. * pi / (l * (l + 1.))
        cltt_theo(l) = Theory%Cls(1,1)%Cl(l) * lfact
        clte_theo(l) = Theory%Cls(2,1)%Cl(l) * lfact
        clee_theo(l) = Theory%Cls(2,2)%Cl(l) * lfact
        if (this%do_delensing) then
          clpp_theo(l) = Theory%Cls(4,4)%Cl(l) * lfact
          ells(l) = l
        end if
      end do

      ! get delensed theory if it's needed
      if (this%do_delensing) then
        write(*,*) "rank ", this%rank, ": calculating delensed theory..."
        cls_size = size(cltt_theo)
        filt_size = size(NL)
!$omp barrier
!$omp master
        call get_delensed(this%rank, cls_size, filt_size, c_loc(cltt_theo), c_loc(clte_theo), c_loc(clee_theo), c_loc(ells), c_loc(clpp_theo), c_loc(L_vals), c_loc(NL), this%kappa_lmin, this%kappa_lmax, c_loc(tt_delen), c_loc(te_delen), c_loc(ee_delen))
!$omp end master
!$omp barrier
        write(*,*) "rank ", this%rank, ": got delensed theory."
      end if

      ! make the cls vector to pass to the likelihood
      allocate(cls_vector(2:this%tt_lmax, 1:3))
      if (this%do_delensing) then
        cls_vector(2:this%tt_lmax,1) = tt_delen(2:this%tt_lmax)
        cls_vector(2:this%tt_lmax,2) = te_delen(2:this%tt_lmax)
        cls_vector(2:this%tt_lmax,3) = ee_delen(2:this%tt_lmax)
      else
        cls_vector(2:this%tt_lmax,1) = cltt_theo(2:this%tt_lmax)
        cls_vector(2:this%tt_lmax,2) = clte_theo(2:this%tt_lmax)
        cls_vector(2:this%tt_lmax,3) = clee_theo(2:this%tt_lmax) 
      end if
      
      ! get values of foreground and actpol params 
      a_tsz = DataParams(1)
      a_ksz = DataParams(2)
      xi = DataParams(3)
      a_d = DataParams(4)
      a_s = DataParams(5)
      a_c = DataParams(6)
      a_g = DataParams(7)
      a_tps = DataParams(8)
      a_ps = DataParams(9)
      a_gte = DataParams(10)
      a_gee = DataParams(11)
      a_ste = DataParams(12)
      a_see = DataParams(13)
      beta_c = DataParams(14)
      ct1 = DataParams(15)
      ct2 = DataParams(16)
      yp1 = DataParams(17)
      yp2 = DataParams(18)
    

      ! call the likelihood function
      actpol_like = 0.0d0
      call actpol_likelihood_compute(cls_vector, a_tsz, a_ksz, xi, a_d, a_s, a_c, a_g, a_tps, a_ps, a_gte, a_gee, a_ste, a_see, beta_c, ct1, ct2, yp1, yp2, this%tt_lmax, this%nbint, this%nbintt, this%nbinte, this%nbinee, this%lmax_win, this%bmax0, this%b1tt, this%b1te, this%b1ee, this%use_cls, this%use_tt, this%use_te, this%use_ee, this%save_test_output, this%test_save_dir, this%test_save_root, this%invcov, this%data_run, actpol_like)
      ACTPol_LnLike = actpol_like
     
      ! print out likelihood valie
      write(*,*) "rank ", this%rank, ": ACTPol like = ", actpol_like

      if(this%save_test_output) then
        write(*,*) "ACTPol like = ", actpol_like
        write(*,*) "Expected like = ", this%expected_like
        write(*,*) "difference = ", actpol_like-this%expected_like
        if (abs((actpol_like-this%expected_like)/this%expected_like) .le. 0.05) then
          write(*,*) "Success! Likelihood matches expected result within 5%."
        else
          write(*,*) "The likelihoods did not match."
        end if
        write(*,*) "fractional difference = ", ((actpol_like-this%expected_like)/this%expected_like) 
      end if
    

      ! deallocate arrays
      deallocate(cltt_theo)
      deallocate(clte_theo)
      deallocate(clee_theo)
      deallocate(cls_vector)
      if (this%do_delensing) then
        deallocate(clpp_theo)
        deallocate(ells)
        deallocate(tt_delen)
        deallocate(te_delen)
        deallocate(ee_delen)
      end if

    write(*,*) "actpol_likelihood.f90: done"

    end function ACTPol_LnLike


end module actpol_likelihood
